import csv
import os
import random
from collections import defaultdict
from io import StringIO
from typing import Dict, List, Union

import sqlalchemy as sa
from aiohttp import ClientSession, ClientTimeout, TCPConnector
from pg_data_flow import DBLoader, as_table
from pydantic import BaseSettings
from ready_logger import get_logger
from sqlalchemy.engine import Engine
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy.sql.elements import quoted_name

from .tables import blacklist

USER_AGENTS = [
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.54 Safari/537.36",
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:92.0) Gecko/20100101 Firefox/92.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.71 Safari/537.36",
]


class S3Settings(BaseSettings):
    s3_endpoint_url: str = "http://localhost:9000"
    aws_access_key_id: str
    aws_secret_access_key: str


def get_postgres_url():
    var = "POSTGRES_URL"
    if var not in os.environ:
        raise ValueError(
            f"Environmental variable {var} is not set. Can not connect to Postgresql."
        )
    return os.environ[var]


def engine_from_env(env_var: str = "POSTGRES_URL") -> Engine:
    """Create an SQLAlchemy engine and chache it so we only create it once.

    Returns:
        Engine: The sqlalchemy engine.
    """
    if not (url := os.getenv(env_var)):
        raise RuntimeError(
            f"Environment variable {env_var} is not set. Can not create engine."
        )
    return sa.create_engine(url)


def get_redis_url():
    var = "REDIS_URL"
    if var not in os.environ:
        raise ValueError(
            f"Environmental variable {var} is not set. Can not connect to Redis."
        )
    return os.environ[var]


def create_http_client():
    return ClientSession(
        timeout=ClientTimeout(60_000), connector=TCPConnector(ssl=False)
    )


def load_csv_with_header(file_text: str) -> List[Dict[str, str]]:
    fo = StringIO(file_text)
    header = [c.strip() for c in next(csv.reader(fo))]
    return list(csv.DictReader(fo, fieldnames=header))


def random_user() -> str:
    return random.choice(USER_AGENTS)


def user_header() -> Dict[str, str]:
    return {"User-Agent": "Danklabs.org. Open Source Data."}  # random_user()}


def extract_column_name(sa_obj: Union[sa.Table, DeclarativeMeta]) -> quoted_name:
    table = as_table(sa_obj)
    column_names = table.columns.keys()
    if len(column_names) > 1:
        raise RuntimeError(
            "Entity %s has more than one column. (%s columns)",
            sa_obj,
            len(column_names),
        )
    return column_names[0]


class BlackLister:
    logger = get_logger("BlackLister")

    @classmethod
    async def create(
        cls, max_bad_count: int = 4, max_bad_ratio: float = 0.7, **upserter_kwargs
    ) -> None:
        self = cls()
        self.max_bad_count = max_bad_count
        self.max_bad_ratio = max_bad_ratio
        self._total = defaultdict(int)
        self._bad = defaultdict(int)
        self._upserter = await DBLoader.create(
            sa_obj=blacklist, on_duplicate_key_update=False, **upserter_kwargs
        )
        return self

    def mark_ok(self, endpoint: str, symbol: str):
        self._total[(endpoint, symbol)] += 1

    async def mark_bad(self, endpoint: str, symbol: str) -> True:
        key = (endpoint, symbol)
        self._bad[key] += 1
        self._total[key] += 1
        if (
            bad_count := self._bad[key]
        ) >= self.max_bad_count and bad_count / self._total[key] >= self.max_bad_ratio:
            self.logger.warning(f"Blacklisting symbol: {symbol}, endpoint: {endpoint}")
            await self._upserter.load([{"endpoint": endpoint, "symbol": symbol}])
            return True
        return False
